@@ -1,7 +1,13 @@
 #!/usr/bin/python3.7
 # -*-coding:utf8 -*
 
+"""
+Module for UnivariateFunctionalData classes.
+
+This module is used define univariate functional data.
+"""
 import itertools
+import inspect
 import numpy as np
 import pygam
 
@@ -21,16 +27,17 @@ def _check_argvals(argvals):
     """Check the user provided `argvals`.
 
     Parameters
-    ---------
-    argvals : list of numpy.ndarray
+    ----------
+    argvals: list of numpy.ndarray
         A list of numeric vectors (numpy.ndarray) or a single numeric vector
         (numpy.ndarray) giving the sampling points in the domains.
 
-    Return
-    ------
-    argvals : list of numpy.ndarray
+    Returns
+    -------
+    list of numpy.ndarray
+
     """
-    if type(argvals) not in (np.ndarray, list):
+    if not isinstance(argvals, (np.ndarray, list)):
         raise ValueError(
             'argvals has to be a list of numpy.ndarray or a numpy.ndarray!')
     if isinstance(argvals, list) and \
@@ -42,7 +49,7 @@ def _check_argvals(argvals):
 
     # Check if all entries of `argvals` are numeric.
     argvals_ = list(itertools.chain.from_iterable(argvals))
-    if not all([type(i) in (int, float, np.int_, np.float_)
+    if not all([isinstance(i, (int, float, np.int_, np.float_))
                 for i in argvals_]):
         raise ValueError(
             'All argvals elements must be numeric!')
@@ -58,11 +65,11 @@ def _check_values(values):
     values : numpy.array
         A numpy array containing values.
 
-    Return
-    ------
+    Returns
+    -------
     values : numpy array
-    """
 
+    """
     # TODO: Modify the function to deal with other types of data.
     if not isinstance(values, np.ndarray):
         raise ValueError(
@@ -81,11 +88,11 @@ def _check_argvals_values(argvals, values):
     values : numpy.ndarray
         Numpy array containing the values.
 
-    Return
-    ------
+    Returns
+    -------
     True, if the argvals and the values are ok.
-    """
 
+    """
     if len(argvals) != len(values.shape[1:]):
         raise ValueError(
             'argvals and values elements have different support dimensions!')
@@ -102,11 +109,14 @@ def _check_argvals_equality(argvals1, argvals2):
     Parameters
     ----------
     argvals1 : list of tuples
+        First argument
     argvals2 : list of tuples
+        Second argument
 
-    Return
-    ------
+    Returns
+    -------
     True, if the two argvals are the same.
+
     """
     if argvals1 != argvals2:
         raise ValueError(
@@ -120,37 +130,29 @@ def _check_argvals_equality(argvals1, argvals2):
 
 
 class UnivariateFunctionalData(object):
-    """An object for defining Univariate Functional Data.
+    """An object for defining Univariate Functional Data."""
 
-    Parameters
-    ----------
-    argvals : list of numpy.ndarray
-        A list of numeric vectors (numpy.ndarray) or a single numeric
-        vector (numpy.ndarray) giving the sampling points in the domains.
-
-    values : array-like
-        An array, giving the observed values for N observations. Missing values
-        should be included via `None` (or `np.nan`). The shape depends
-        on `argvals`::
-
-            (N, M) if `argvals` is a single numeric vector,
-            (N, M_1, ..., M_d) if `argvals` is a list of numeric vectors.
-
-    standardize : boolean, default = True
-        Do we standardize the argvals to be in [0, 1].
-
-    Attributes
-    ----------
-
-    Notes
-    -----
-
-    References
-    ----------
-
-    """
     def __init__(self, argvals, values, standardize=True):
+        """Initialize UnivariateFunctionalData object.
 
+        Parameters
+        ----------
+        argvals : list of numpy.ndarray
+            A list of numeric vectors (numpy.ndarray) or a single numeric
+            vector (numpy.ndarray) giving the sampling points in the domains.
+
+        values : array-like
+            An array, giving the observed values for N observations.
+            Missing values should be included via `None` (or `np.nan`). The
+            shape depends on `argvals`::
+
+                (N, M) if `argvals` is a single numeric vector,
+                (N, M_1, ..., M_d) if `argvals` is a list of numeric vectors.
+
+        standardize : boolean, default = True
+            Do we standardize the argvals to be in [0, 1].
+
+        """
         self.argvals = argvals
         self.values = values
 
@@ -161,6 +163,7 @@ class UnivariateFunctionalData(object):
             self.argvals_stand = argvals_stand
 
     def __repr__(self):
+        """Override print function."""
         res = "Univariate Functional data objects with " +\
             str(self.nObs()) +\
             " observations of " +\
@@ -182,15 +185,15 @@ class UnivariateFunctionalData(object):
         return res
 
     def __getitem__(self, index):
-        """Function call when self[index]
+        """Function call when self[index].
 
         Parameters
         ----------
         index : int
             The observation(s) of the object to retrieve.
 
-        Return
-        ------
+        Returns
+        -------
         res : UnivariateFunctionalData object
             The selected observation(s) as UnivariateFunctionalData object.
 
@@ -205,6 +208,7 @@ class UnivariateFunctionalData(object):
         return res
 
     def __add__(self, new):
+        """Override add function."""
         if not isinstance(new, UnivariateFunctionalData):
             raise ValueError(
                 """The object to add must be an object of the class
@@ -215,6 +219,7 @@ class UnivariateFunctionalData(object):
         return res
 
     def __sub__(self, new):
+        """Override sub funcion."""
         if not isinstance(new, UnivariateFunctionalData):
             raise ValueError(
                 """The object to substract must be an object of the class
@@ -225,8 +230,9 @@ class UnivariateFunctionalData(object):
         return res
 
     def __mul__(self, obj):
+        """Overide mul function."""
         values = np.empty(shape=self.values.shape)
-        if type(obj) in (int, float, np.int_, np.float_):
+        if isinstance(obj, (int, float, np.int_, np.float_)):
             values = self.values * obj
         elif (isinstance(obj, list)) and (self.nObs() == len(obj)):
             for i in np.arange(0, len(obj)):
@@ -240,10 +246,12 @@ class UnivariateFunctionalData(object):
         return res
 
     def __rmul__(self, nb):
+        """Override rmul function."""
         return self * nb
 
     @property
     def argvals(self):
+        """Getter for argvals."""
         return self._argvals
 
     @argvals.setter
@@ -255,6 +263,7 @@ class UnivariateFunctionalData(object):
 
     @property
     def argvals_stand(self):
+        """Getter for argvals_stand."""
         return self._argvals_stand
 
     @argvals_stand.setter
@@ -263,6 +272,7 @@ class UnivariateFunctionalData(object):
 
     @property
     def values(self):
+        """Getter for values."""
         return self._values
 
     @values.setter
@@ -274,6 +284,7 @@ class UnivariateFunctionalData(object):
 
     @property
     def mean_(self):
+        """Getter for mean_."""
         return self._mean_
 
     @mean_.setter
@@ -283,8 +294,8 @@ class UnivariateFunctionalData(object):
     def nObs(self):
         """Number of observations of the object.
 
-        Return
-        ------
+        Returns
+        -------
         n : int
             Number of observations of the object.
 
@@ -295,8 +306,8 @@ class UnivariateFunctionalData(object):
     def rangeObs(self):
         """Range of the observations of the objects.
 
-        Return
-        ------
+        Returns
+        -------
         min(values_), max(values_) : tuple
             Tuple containing the minimum and maximum number of all the
             observations for an object.
@@ -315,8 +326,8 @@ class UnivariateFunctionalData(object):
     def nObsPoint(self):
         """Number of sampling points of the objects.
 
-        Return
-        ------
+        Returns
+        -------
         n : list of int
             List of the length self.dimension() where the i-th entry
             correspond to the number of sampling points of the i-th dimension
@@ -329,11 +340,12 @@ class UnivariateFunctionalData(object):
     def rangeObsPoint(self):
         """Range of the observations of the objects.
 
-        Return
-        ------
+        Returns
+        -------
         range_ : list of tuples containing the minimum and maximum number
         where the i-th entry of the list contains the range of the i-th
         dimension of the object.
+
         """
         range_ = [(min(i), max(i)) for i in self.argvals]
         return range_
@@ -341,8 +353,8 @@ class UnivariateFunctionalData(object):
     def dimension(self):
         """Common dimension of the observations of the object.
 
-        Return
-        ------
+        Returns
+        -------
         dim : int
             Number of dimension of the observations of the object.
 
@@ -351,15 +363,17 @@ class UnivariateFunctionalData(object):
         return dim
 
     def asIrregularFunctionalData(self):
-        """Coerce univariate functional data of dimension 1 into irregular
+        """Convert to univariate to irregular functional data.
+
+        Coerce univariate functional data of dimension 1 into irregular
         functional data.
 
-        Return
-        ------
+        Returns
+        -------
         obj : IrregularFunctionalData
             An object of the class IrregularFunctionalData
-        """
 
+        """
         if self.dimension() != 1:
             raise ValueError(
                 """It is not possible to coerce a UnivariateFunctionalData as
@@ -396,8 +410,8 @@ class UnivariateFunctionalData(object):
                 - n_basis: int
                     default=10
 
-        Return
-        ------
+        Returns
+        -------
         obj : UnivariateFunctionalData object
             Object of the class UnivariateFunctionalData with the same
             argvals as self and one observation.
@@ -405,7 +419,7 @@ class UnivariateFunctionalData(object):
         """
         mean_ = rowMean_(self.values)
         if smooth:
-            if method is 'LocalLinear':
+            if method == 'LocalLinear':
                 kernel = kwargs.get('kernel', 'gaussian')
                 degree = kwargs.get('degree', 2)
                 bandwidth = kwargs.get('bandwidth', 1)
@@ -415,7 +429,7 @@ class UnivariateFunctionalData(object):
                                      degree=degree)
                 lp.fit(self.argvals, mean_)
                 mean_ = lp.X_fit_
-            elif method is 'GAM':
+            elif method == 'GAM':
                 n_basis = kwargs.get('n_basis', 10)
 
                 X = np.array(self.argvals[0])
@@ -451,8 +465,8 @@ class UnivariateFunctionalData(object):
                 - n_basis: int
                     default=10
 
-        Return
-        ------
+        Returns
+        -------
         obj : UnivariateFunctionalData object
             Object of the class UnivariateFunctionalData with dimension 2
             and one observation.
@@ -467,6 +481,7 @@ class UnivariateFunctionalData(object):
         -----
         Currently, for smoothing, please consider to use the 'GAM' method as
         the 'LocalLinear' is not working.
+
         """
         if self.dimension() != 1:
             raise ValueError(
@@ -480,7 +495,7 @@ class UnivariateFunctionalData(object):
         diag = np.copy(np.diag(cov))
 
         if smooth:
-            if method is 'LocalLinear':
+            if method == 'LocalLinear':
                 kernel = kwargs.get('kernel', 'gaussian')
                 degree = kwargs.get('degree', 2)
                 bandwidth = kwargs.get('bandwidth', 1)
@@ -490,7 +505,7 @@ class UnivariateFunctionalData(object):
                                      degree=degree)
                 lp.fit(new_argvals, cov)
                 cov = lp.X_fit_
-            elif method is 'GAM':
+            elif method == 'GAM':
                 n_basis = kwargs.get('n_basis', 10)
 
                 # Remove covariance diagonale because of measurement errors.
@@ -532,9 +547,9 @@ class UnivariateFunctionalData(object):
         T1_lower = np.sum(~(D >= (D[0] + 0.25 * T_len)))
         T1_upper = np.sum((D <= (D[len(D) - 1] - 0.25 * T_len)))
         W = integrationWeights_(D[T1_lower:T1_upper], method='trapz')
-        sigma2 = np.maximum(
-            np.dot(W, (V_hat - np.diag(cov))[T1_lower:T1_upper]) /
-            (D[T1_upper] - D[T1_lower]), 0)
+
+        nume = np.dot(W, (V_hat - np.diag(cov))[T1_lower:T1_upper])
+        sigma2 = np.maximum(nume / (D[T1_upper] - D[T1_lower]), 0)
 
         self.covariance_ = UnivariateFunctionalData(new_argvals,
                                                     np.array(cov, ndmin=3))
@@ -545,11 +560,12 @@ class UnivariateFunctionalData(object):
 
         This method estimates the (heteroscedastic) noise for a univariate
         functional data object.
-        Model: :math:`Z_i(t_k) = f_i(t_k) + \sigma(f_i(t_k))\epsilon_i`
+        Model: :math:`Z_i(t_k) = f_i(t_k) + sigma(f_i(t_k))epsilon_i`
         It is assume that all the curves have been sampled on the same design
         points. Let's t_1, ..., t_k be that points, the estimation of the
         noise at t_k is::
-        :math:`sigma^2(t_k) = 1/2n \sum_{i} [Z_i(t_{k+1}) - Z_i(t_{k})]^2`
+        :math:`sigma^2(t_k) = 1/2n sum_{i} [Z_i(t_{k+1}) - Z_i(t_{k})]^2`
+
         """
         if self.dimension() != 1:
             raise ValueError(
@@ -564,8 +580,8 @@ class UnivariateFunctionalData(object):
             The method used to integrated. Currently, only the Simpsons method
             is implemented.
 
-        Return
-        ------
+        Returns
+        -------
         obj : list of int
             List where entry i is the integration of the observation i over
             the argvals.
@@ -573,8 +589,9 @@ class UnivariateFunctionalData(object):
         Note
         ----
         Only work with 1-dimensional functional data.
+
         """
-        if method is not 'simpson':
+        if method != 'simpson':
             raise ValueError('Only the Simpsons method is implemented!')
         return [integrate_(self.argvals[0], i) for i in self.values]
 
@@ -586,11 +603,12 @@ class UnivariateFunctionalData(object):
         data : FDApy.univariate_functional.UnivariateFunctionalData object
             A one dimensional univariate functional data
 
-        Return
-        ------
+        Returns
+        -------
         obj : FDApy.univariate_functional.UnivariateFunctionalData
             Object of the class UnivariateFunctionalData of dimension 2
             (self.argvals x data.argvals)
+
         """
         if (self.dimension() != 1) or (data.dimension() != 1):
             raise ValueError(
@@ -601,8 +619,8 @@ class UnivariateFunctionalData(object):
                       for j in data.values]
         return UnivariateFunctionalData(new_argvals, np.array(new_values))
 
-    def smooth(self, t0, k0,
-               points=None, degree=0, kernel="epanechnikov", bandwidth=None):
+    def smooth(self, t0, k0, points=None, degree=0, kernel="epanechnikov",
+               bandwidth=None):
         """Smooth the data.
 
         Currently, it uses local polynomial regression.
@@ -620,11 +638,12 @@ class UnivariateFunctionalData(object):
         degree: integer, default=2
             Degree of the local polynomial to fit.
 
-        Return
-        ------
+        Returns
+        -------
         res : FDApy.univariate_functional.UnivariateFunctionalData
             Object of the class UnivariateFunctionalData which correpond to
             the data that have been smooth:: argvals = `points` given as input
+
         """
         if self.dimension() != 1:
             raise ValueError(
@@ -637,3 +656,59 @@ class UnivariateFunctionalData(object):
                                   kernel=kernel, bandwidth=bandwidth)
 
         return data_smooth
+
+    def add_noise(self, noise_var=1):
+        r"""Add noise to the data.
+
+        Parameters
+        ----------
+        noise_var : float or callable, default = 1
+            Variance of the noise to add. May be callable for heteroscedastic
+            noise.
+
+        Notes
+        -----
+
+        Model:
+
+        .. math::
+            Z(t) = f(t) + \sigma(f(t))\epsilon
+
+        """
+        noisy_data = np.random.normal(0, 1, (self.nObs(), self.nObsPoint()[0]))
+
+        if inspect.isfunction(noise_var):
+            noise_var = noise_var(self.values)
+
+        noise_std = np.sqrt(noise_var)
+        noisy_data = self.values + np.multiply(noise_std, noisy_data)
+        return UnivariateFunctionalData(self.argvals, noisy_data)
+
+    def sparsify(self, perc=0.9, epsilon=0.05):
+        """Sparsify the data.
+
+        Arguments
+        ---------
+        perc: double, default = 0.9
+            Percentage of data to keep.
+        epsilon: double, default = 0.05
+            Uncertainty on the percentage to keep.
+
+        Returns
+        -------
+        obj: IrregularFunctionalData
+            The data that have been sparsify.
+
+        """
+        argvals = []
+        values = []
+        for obs in self:
+            s = obs.values.size
+            p = np.random.uniform(max(0, perc - epsilon),
+                                  min(1, perc + epsilon))
+            idx = np.sort(np.random.choice(np.arange(0, s),
+                                           size=int(p * s),
+                                           replace=False))
+            argvals.append(obs.argvals[0][idx])
+            values.append(obs.values[0][idx])
+        return IrregularFunctionalData(argvals, values)
